"""empty message

Revision ID: fb5bd88ae2f3
Revises: 93f9354fa923
Create Date: 2020-05-14 20:00:56.474904

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "fb5bd88ae2f3"
down_revision = "4b27843a188a"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("Create Index ix_task_waiting_sort on task_queue (priority desc, created_on)")
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade():
    op.execute("Drop Index ix_task_waiting_sort")
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###
